#!/usr/bin/env python3

# import logging

# logging.basicConfig(
#     level=logging.INFO,
#     format='%(asctime)s %(levelname)s:%(name)s:%(message)s',  # See https://docs.python.org/3/library/logging.html#logrecord-attributes
#     datefmt=r'%Y-%m-%dT%H:%M:%S%z'
# )

def main():
    pass

if __name__ == "__main__":
    main()
